@@ -17,16 +17,6 @@
 #include "sqliteInt.h"
 #include "os.h"
 
-#ifndef OS_UNIX
-# ifndef OS_WIN
-#  define OS_UNIX 1
-# else
-#  define OS_UNIX 0
-# endif
-#endif
-#ifndef OS_WIN
-# define OS_WIN 0
-#endif
 #if OS_UNIX
 # include <unistd.h>
 # include <fcntl.h>
@@ -37,6 +27,20 @@
 # include <winbase.h>
 #endif
 
+/*
+** Macros for performance tracing.  Normally turned off
+*/
+#if 0
+static int last_page = 0;
+#define SEEK(X)     last_page=(X)
+#define TRACE1(X)   fprintf(stderr,X)
+#define TRACE2(X,Y) fprintf(stderr,X,Y)
+#else
+#define SEEK(X)
+#define TRACE1(X)
+#define TRACE2(X,Y)
+#endif
+
 
 #if OS_UNIX
 /*
@@ -108,7 +112,7 @@ struct inodeKey {
 */
 struct lockInfo {
   struct inodeKey key;  /* The lookup key */
-  int cnt;              /* 0: unlocked.  -1: write lock.  >=1: read lock */
+  int cnt;              /* 0: unlocked.  -1: write lock.  1...: read lock. */
   int nRef;             /* Number of pointers to this structure */
 };
 
@@ -130,16 +134,23 @@ static struct lockInfo *findLockInfo(int fd){
   struct lockInfo *pInfo;
   rc = fstat(fd, &statbuf);
   if( rc!=0 ) return 0;
+  memset(&key, 0, sizeof(key));
   key.dev = statbuf.st_dev;
   key.ino = statbuf.st_ino;
   pInfo = (struct lockInfo*)sqliteHashFind(&lockHash, &key, sizeof(key));
   if( pInfo==0 ){
+    struct lockInfo *pOld;
     pInfo = sqliteMalloc( sizeof(*pInfo) );
     if( pInfo==0 ) return 0;
     pInfo->key = key;
     pInfo->nRef = 1;
     pInfo->cnt = 0;
-    sqliteHashInsert(&lockHash, &pInfo->key, sizeof(key), pInfo);
+    pOld = sqliteHashInsert(&lockHash, &pInfo->key, sizeof(key), pInfo);
+    if( pOld!=0 ){
+      assert( pOld==pInfo );
+      sqliteFree(pInfo);
+      pInfo = 0;
+    }
   }else{
     pInfo->nRef++;
   }
@@ -197,20 +208,7 @@ int sqliteOsFileExists(const char *zFilename){
   return access(zFilename, 0)==0;
 #endif
 #if OS_WIN
-  HANDLE h;
-  h = CreateFile(zFilename,
-    GENERIC_READ,
-    0,
-    NULL,
-    OPEN_EXISTING,
-    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
-    NULL
-  );
-  if( h!=INVALID_HANDLE_VALUE ){
-    CloseHandle(h);
-    return 1;
-  }
-  return 0;
+  return GetFileAttributes(zFilename) != 0xffffffff;
 #endif
 }
 
@@ -220,7 +218,7 @@ int sqliteOsFileExists(const char *zFilename){
 ** fails, try opening it read-only.  If the file does not exist,
 ** try to create it.
 **
-** On success, a handle for the open file is written to *pResult
+** On success, a handle for the open file is written to *id
 ** and *pReadonly is set to 0 if the file was opened for reading and
 ** writing or 1 if the file was opened read-only.  The function returns
 ** SQLITE_OK.
@@ -230,15 +228,14 @@ int sqliteOsFileExists(const char *zFilename){
 */
 int sqliteOsOpenReadWrite(
   const char *zFilename,
-  OsFile *pResult,
+  OsFile *id,
   int *pReadonly
 ){
 #if OS_UNIX
-  OsFile s;
-  s.fd = open(zFilename, O_RDWR|O_CREAT, 0644);
-  if( s.fd<0 ){
-    s.fd = open(zFilename, O_RDONLY);
-    if( s.fd<0 ){
+  id->fd = open(zFilename, O_RDWR|O_CREAT, 0644);
+  if( id->fd<0 ){
+    id->fd = open(zFilename, O_RDONLY);
+    if( id->fd<0 ){
       return SQLITE_CANTOPEN; 
     }
     *pReadonly = 1;
@@ -246,13 +243,13 @@ int sqliteOsOpenReadWrite(
     *pReadonly = 0;
   }
   sqliteOsEnterMutex();
-  s.pLock = findLockInfo(s.fd);
+  id->pLock = findLockInfo(id->fd);
   sqliteOsLeaveMutex();
-  if( s.pLock==0 ){
-    close(s.fd);
+  if( id->pLock==0 ){
+    close(id->fd);
     return SQLITE_NOMEM;
   }
-  *pResult = s;
+  id->locked = 0;
   return SQLITE_OK;
 #endif
 #if OS_WIN
@@ -280,7 +277,8 @@ int sqliteOsOpenReadWrite(
   }else{
     *pReadonly = 0;
   }
-  *pResult = h;
+  id->h = h;
+  id->locked = 0;
   return SQLITE_OK;
 #endif
 }
@@ -293,46 +291,61 @@ int sqliteOsOpenReadWrite(
 ** previously existed.  Nor do we allow the file to be a symbolic
 ** link.
 **
-** On success, write the file handle into *pResult and return SQLITE_OK.
+** If delFlag is true, then make arrangements to automatically delete
+** the file when it is closed.
+**
+** On success, write the file handle into *id and return SQLITE_OK.
 **
 ** On failure, return SQLITE_CANTOPEN.
 */
-int sqliteOsOpenExclusive(const char *zFilename, OsFile *pResult){
+int sqliteOsOpenExclusive(const char *zFilename, OsFile *id, int delFlag){
 #if OS_UNIX
-  OsFile s;
   if( access(zFilename, 0)==0 ){
     return SQLITE_CANTOPEN;
   }
 #ifndef O_NOFOLLOW
 # define O_NOFOLLOW 0
 #endif
-  s.fd = open(zFilename, O_RDWR|O_CREAT|O_EXCL|O_NOFOLLOW, 0600);
-  if( s.fd<0 ){
+  id->fd = open(zFilename, O_RDWR|O_CREAT|O_EXCL|O_NOFOLLOW, 0600);
+  if( id->fd<0 ){
     return SQLITE_CANTOPEN;
   }
   sqliteOsEnterMutex();
-  s.pLock = findLockInfo(s.fd);
+  id->pLock = findLockInfo(id->fd);
   sqliteOsLeaveMutex();
-  if( s.pLock==0 ){
-    close(s.fd);
+  if( id->pLock==0 ){
+    close(id->fd);
+    unlink(zFilename);
     return SQLITE_NOMEM;
   }
-  *pResult = s;
+  id->locked = 0;
+  if( delFlag ){
+    unlink(zFilename);
+  }
   return SQLITE_OK;
 #endif
 #if OS_WIN
-  HANDLE h = CreateFile(zFilename,
+  HANDLE h;
+  int fileflags;
+  if( delFlag ){
+    fileflags = FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_RANDOM_ACCESS 
+                     | FILE_FLAG_DELETE_ON_CLOSE;
+  }else{
+    fileflags = FILE_FLAG_RANDOM_ACCESS;
+  }
+  h = CreateFile(zFilename,
      GENERIC_READ | GENERIC_WRITE,
      0,
      NULL,
      CREATE_ALWAYS,
-     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
+     fileflags,
      NULL
   );
   if( h==INVALID_HANDLE_VALUE ){
     return SQLITE_CANTOPEN;
   }
-  *pResult = h;
+  id->h = h;
+  id->locked = 0;
   return SQLITE_OK;
 #endif
 }
@@ -340,25 +353,24 @@ int sqliteOsOpenExclusive(const char *zFilename, OsFile *pResult){
 /*
 ** Attempt to open a new file for read-only access.
 **
-** On success, write the file handle into *pResult and return SQLITE_OK.
+** On success, write the file handle into *id and return SQLITE_OK.
 **
 ** On failure, return SQLITE_CANTOPEN.
 */
-int sqliteOsOpenReadOnly(const char *zFilename, OsFile *pResult){
+int sqliteOsOpenReadOnly(const char *zFilename, OsFile *id){
 #if OS_UNIX
-  OsFile s;
-  s.fd = open(zFilename, O_RDONLY);
-  if( s.fd<0 ){
+  id->fd = open(zFilename, O_RDONLY);
+  if( id->fd<0 ){
     return SQLITE_CANTOPEN;
   }
   sqliteOsEnterMutex();
-  s.pLock = findLockInfo(s.fd);
+  id->pLock = findLockInfo(id->fd);
   sqliteOsLeaveMutex();
-  if( s.pLock==0 ){
-    close(s.fd);
+  if( id->pLock==0 ){
+    close(id->fd);
     return SQLITE_NOMEM;
   }
-  *pResult = s;
+  id->locked = 0;
   return SQLITE_OK;
 #endif
 #if OS_WIN
@@ -373,7 +385,8 @@ int sqliteOsOpenReadOnly(const char *zFilename, OsFile *pResult){
   if( h==INVALID_HANDLE_VALUE ){
     return SQLITE_CANTOPEN;
   }
-  *pResult = h;
+  id->h = h;
+  id->locked = 0;
   return SQLITE_OK;
 #endif
 }
@@ -385,10 +398,10 @@ int sqliteOsOpenReadOnly(const char *zFilename, OsFile *pResult){
 int sqliteOsTempFileName(char *zBuf){
 #if OS_UNIX
   static const char *azDirs[] = {
-     ".",
      "/var/tmp",
      "/usr/tmp",
      "/tmp",
+     ".",
   };
   static char zChars[] =
     "abcdefghijklmnopqrstuvwxyz"
@@ -422,8 +435,10 @@ int sqliteOsTempFileName(char *zBuf){
   int i, j;
   char zTempPath[SQLITE_TEMPNAME_SIZE];
   GetTempPath(SQLITE_TEMPNAME_SIZE-30, zTempPath);
+  for(i=strlen(zTempPath); i>0 && zTempPath[i-1]=='\\'; i--){}
+  zTempPath[i] = 0;
   for(;;){
-    sprintf(zBuf, "%s/sqlite_", zTempPath);
+    sprintf(zBuf, "%s\\sqlite_", zTempPath);
     j = strlen(zBuf);
     for(i=0; i<15; i++){
       int n = sqliteRandomByte() % sizeof(zChars);
@@ -439,36 +454,38 @@ int sqliteOsTempFileName(char *zBuf){
 /*
 ** Close a file
 */
-int sqliteOsClose(OsFile id){
+int sqliteOsClose(OsFile *id){
 #if OS_UNIX
-  close(id.fd);
+  close(id->fd);
   sqliteOsEnterMutex();
-  releaseLockInfo(id.pLock);
+  releaseLockInfo(id->pLock);
   sqliteOsLeaveMutex();
   return SQLITE_OK;
 #endif
 #if OS_WIN
-  CloseHandle(id);
+  CloseHandle(id->h);
   return SQLITE_OK;
 #endif
 }
 
 /*
-** Read data from a file into a buffer.  Return the number of
-** bytes actually read.
+** Read data from a file into a buffer.  Return SQLITE_OK if all
+** bytes were read successfully and SQLITE_IOERR if anything goes
+** wrong.
 */
-int sqliteOsRead(OsFile id, void *pBuf, int amt){
+int sqliteOsRead(OsFile *id, void *pBuf, int amt){
 #if OS_UNIX
   int got;
   SimulateIOError(SQLITE_IOERR);
-  got = read(id.fd, pBuf, amt);
+  TRACE2("READ %d\n", last_page);
+  got = read(id->fd, pBuf, amt);
   if( got<0 ) got = 0;
   return got==amt ? SQLITE_OK : SQLITE_IOERR;
 #endif
 #if OS_WIN
   DWORD got;
   SimulateIOError(SQLITE_IOERR);
-  if( !ReadFile(id, pBuf, amt, &got, 0) ){
+  if( !ReadFile(id->h, pBuf, amt, &got, 0) ){
     got = 0;
   }
   return got==amt ? SQLITE_OK : SQLITE_IOERR;
@@ -479,18 +496,19 @@ int sqliteOsRead(OsFile id, void *pBuf, int amt){
 ** Write data from a buffer into a file.  Return SQLITE_OK on success
 ** or some other error code on failure.
 */
-int sqliteOsWrite(OsFile id, const void *pBuf, int amt){
+int sqliteOsWrite(OsFile *id, const void *pBuf, int amt){
 #if OS_UNIX
   int wrote;
   SimulateIOError(SQLITE_IOERR);
-  wrote = write(id.fd, pBuf, amt);
+  TRACE2("WRITE %d\n", last_page);
+  wrote = write(id->fd, pBuf, amt);
   if( wrote<amt ) return SQLITE_FULL;
   return SQLITE_OK;
 #endif
 #if OS_WIN
   DWORD wrote;
   SimulateIOError(SQLITE_IOERR);
-  if( !WriteFile(id, pBuf, amt, &wrote, 0) || wrote<amt ){
+  if( !WriteFile(id->h, pBuf, amt, &wrote, 0) || (int)wrote<amt ){
     return SQLITE_FULL;
   }
   return SQLITE_OK;
@@ -500,13 +518,14 @@ int sqliteOsWrite(OsFile id, const void *pBuf, int amt){
 /*
 ** Move the read/write pointer in a file.
 */
-int sqliteOsSeek(OsFile id, int offset){
+int sqliteOsSeek(OsFile *id, int offset){
+  SEEK(offset/1024 + 1);
 #if OS_UNIX
-  lseek(id.fd, offset, SEEK_SET);
+  lseek(id->fd, offset, SEEK_SET);
   return SQLITE_OK;
 #endif
 #if OS_WIN
-  SetFilePointer(id, offset, 0, FILE_BEGIN);
+  SetFilePointer(id->h, offset, 0, FILE_BEGIN);
   return SQLITE_OK;
 #endif
 }
@@ -514,27 +533,28 @@ int sqliteOsSeek(OsFile id, int offset){
 /*
 ** Make sure all writes to a particular file are committed to disk.
 */
-int sqliteOsSync(OsFile id){
+int sqliteOsSync(OsFile *id){
   SimulateIOError(SQLITE_IOERR);
+  TRACE1("SYNC\n");
 #if OS_UNIX
-  return fsync(id.fd)==0 ? SQLITE_OK : SQLITE_IOERR;
+  return fsync(id->fd)==0 ? SQLITE_OK : SQLITE_IOERR;
 #endif
 #if OS_WIN
-  return FlushFileBuffers(id) ? SQLITE_OK : SQLITE_IOERR;
+  return FlushFileBuffers(id->h) ? SQLITE_OK : SQLITE_IOERR;
 #endif
 }
 
 /*
 ** Truncate an open file to a specified size
 */
-int sqliteOsTruncate(OsFile id, int nByte){
+int sqliteOsTruncate(OsFile *id, int nByte){
   SimulateIOError(SQLITE_IOERR);
 #if OS_UNIX
-  return ftruncate(id.fd, nByte)==0 ? SQLITE_OK : SQLITE_IOERR;
+  return ftruncate(id->fd, nByte)==0 ? SQLITE_OK : SQLITE_IOERR;
 #endif
 #if OS_WIN
-  SetFilePointer(id, nByte, 0, FILE_BEGIN);
-  SetEndOfFile(id);
+  SetFilePointer(id->h, nByte, 0, FILE_BEGIN);
+  SetEndOfFile(id->h);
   return SQLITE_OK;
 #endif
 }
@@ -542,11 +562,11 @@ int sqliteOsTruncate(OsFile id, int nByte){
 /*
 ** Determine the current size of a file in bytes
 */
-int sqliteOsFileSize(OsFile id, int *pSize){
+int sqliteOsFileSize(OsFile *id, int *pSize){
 #if OS_UNIX
   struct stat buf;
   SimulateIOError(SQLITE_IOERR);
-  if( fstat(id.fd, &buf)!=0 ){
+  if( fstat(id->fd, &buf)!=0 ){
     return SQLITE_IOERR;
   }
   *pSize = buf.st_size;
@@ -554,86 +574,190 @@ int sqliteOsFileSize(OsFile id, int *pSize){
 #endif
 #if OS_WIN
   SimulateIOError(SQLITE_IOERR);
-  *pSize = GetFileSize(id, 0);
+  *pSize = GetFileSize(id->h, 0);
   return SQLITE_OK;
 #endif
 }
 
+/*
+** Windows file locking notes:
+**
+** We cannot use LockFileEx() or UnlockFileEx() because those functions
+** are not available under Win95/98/ME.  So we use only LockFile() and
+** UnlockFile().
+**
+** A read lock is obtained by locking a single random byte in the
+** range of 1 to MX_LOCKBYTE.  The lock byte is obtained at random so
+** two separate readers can probably access the file at the same time,
+** unless they are unlucky and choose the same lock byte.  A write lock
+** is obtained by locking all bytes in the range of 1 to MX_LOCKBYTE.
+** There can only be one writer.
+**
+** A lock is obtained on byte 0 before acquiring either a read lock or
+** a write lock.  This prevents two processes from attempting to get a
+** lock at a same time.  The semantics of sqliteOsReadLock() require that
+** if there is already a write lock, that lock is converted into a read
+** lock atomically.  The lock on byte 0 allows us to drop the old write
+** lock and get the read lock without another process jumping into the
+** middle and messing us up.  The same argument applies to sqliteOsWriteLock().
+**
+** There are a finite number of read locks under windows.  That number
+** is determined by the following variable:
+*/
+#define MX_LOCKBYTE 10240
 
 /*
-** Get a read or write lock on a file.
+** Change the status of the lock on the file "id" to be a readlock.
+** If the file was write locked, then this reduces the lock to a read.
+** If the file was read locked, then this acquires a new read lock.
+**
+** Return SQLITE_OK on success and SQLITE_BUSY on failure.
 */
-int sqliteOsLock(OsFile id, int wrlock){
+int sqliteOsReadLock(OsFile *id){
 #if OS_UNIX
   int rc;
-  int needSysLock;
   sqliteOsEnterMutex();
-  if( wrlock ){
-    if( id.pLock->cnt!=0 ){
-      rc = SQLITE_BUSY;
-    }else{
-      rc = SQLITE_OK;
-      id.pLock->cnt = -1;
-      needSysLock = 1;
+  if( id->pLock->cnt>0 ){
+    if( !id->locked ){
+      id->pLock->cnt++;
+      id->locked = 1;
     }
-  }else{
-    if( id.pLock<0 ){
-      rc = SQLITE_BUSY;
-    }else{
-      rc = SQLITE_OK;
-      needSysLock = id.pLock->cnt==0;
-      id.pLock->cnt++;
-    }
-  }
-  sqliteOsLeaveMutex();      
-  if( rc==SQLITE_OK && needSysLock ){ 
+    rc = SQLITE_OK;
+  }else if( id->locked || id->pLock->cnt==0 ){
     struct flock lock;
-    lock.l_type = wrlock ? F_WRLCK : F_RDLCK;
+    lock.l_type = F_RDLCK;
     lock.l_whence = SEEK_SET;
     lock.l_start = lock.l_len = 0L;
-    rc = fcntl(id.fd, F_SETLK, &lock)==0 ? SQLITE_OK : SQLITE_BUSY;
+    if( fcntl(id->fd, F_SETLK, &lock)!=0 ){
+      rc = SQLITE_BUSY;
+    }else{
+      rc = SQLITE_OK;
+      id->pLock->cnt = 1;
+      id->locked = 1;
+    }
+  }else{
+    rc = SQLITE_BUSY;
   }
+  sqliteOsLeaveMutex();
   return rc;
 #endif
 #if OS_WIN
-  if( !LockFile(id, 0, 0, 1024, 0) ){
-    return SQLITE_BUSY;
+  int rc;
+  if( id->locked>0 ){
+    rc = SQLITE_OK;
+  }else{
+    int lk = (sqliteRandomInteger() & 0x7ffffff)%MX_LOCKBYTE + 1;
+    int res;
+    if( (res = LockFile(id->h, 0, 0, 1, 0))!=0 ){
+      UnlockFile(id->h, 1, 0, MX_LOCKBYTE, 0);
+      res = LockFile(id->h, lk, 0, 1, 0);
+      UnlockFile(id->h, 0, 0, 1, 0);
+    }
+    if( res ){
+      id->locked = lk;
+      rc = SQLITE_OK;
+    }else{
+      rc = SQLITE_BUSY;
+    }
   }
-  return SQLITE_OK;
+  return rc;
 #endif
 }
 
 /*
-** Release the read or write lock from a file.
+** Change the lock status to be an exclusive or write lock.  Return
+** SQLITE_OK on success and SQLITE_BUSY on a failure.
 */
-int sqliteOsUnlock(OsFile id){
+int sqliteOsWriteLock(OsFile *id){
 #if OS_UNIX
   int rc;
-  int needSysUnlock;
-
   sqliteOsEnterMutex();
-  if( id.pLock->cnt<0 ){
-    needSysUnlock = 1;
-    id.pLock->cnt = 0;
-  }else if( id.pLock->cnt>0 ){
-    id.pLock->cnt--;
-    needSysUnlock = id.pLock->cnt==0;
+  if( id->pLock->cnt==0 || (id->pLock->cnt==1 && id->locked==1) ){
+    struct flock lock;
+    lock.l_type = F_WRLCK;
+    lock.l_whence = SEEK_SET;
+    lock.l_start = lock.l_len = 0L;
+    if( fcntl(id->fd, F_SETLK, &lock)!=0 ){
+      rc = SQLITE_BUSY;
+    }else{
+      rc = SQLITE_OK;
+      id->pLock->cnt = -1;
+      id->locked = 1;
+    }
   }else{
-    rc = SQLITE_OK;
-    needSysUnlock = 0;
+    rc = SQLITE_BUSY;
   }
   sqliteOsLeaveMutex();
-  if( needSysUnlock ){
+  return rc;
+#endif
+#if OS_WIN
+  int rc;
+  if( id->locked<0 ){
+    rc = SQLITE_OK;
+  }else{
+    int res;
+    if( (res = LockFile(id->h, 0, 0, 1, 0))!=0 ){
+      if( id->locked==0 || UnlockFile(id->h, id->locked, 0, 1, 0) ){
+        res = LockFile(id->h, 1, 0, MX_LOCKBYTE, 0);
+      }else{
+        res = 0;
+      }
+      UnlockFile(id->h, 0, 0, 1, 0);
+    }
+    if( res ){
+      id->locked = -1;
+      rc = SQLITE_OK;
+    }else{
+      rc = SQLITE_BUSY;
+    }
+  }
+  return rc;
+#endif
+}
+
+/*
+** Unlock the given file descriptor.  If the file descriptor was
+** not previously locked, then this routine is a no-op.
+*/
+int sqliteOsUnlock(OsFile *id){
+#if OS_UNIX
+  int rc;
+  if( !id->locked ) return SQLITE_OK;
+  sqliteOsEnterMutex();
+  assert( id->pLock->cnt!=0 );
+  if( id->pLock->cnt>1 ){
+    id->pLock->cnt--;
+    rc = SQLITE_OK;
+  }else{
     struct flock lock;
     lock.l_type = F_UNLCK;
     lock.l_whence = SEEK_SET;
     lock.l_start = lock.l_len = 0L;
-    rc = fcntl(id.fd, F_SETLK, &lock)==0 ? SQLITE_OK : SQLITE_IOERR;
+    if( fcntl(id->fd, F_SETLK, &lock)!=0 ){
+      rc = SQLITE_BUSY;
+    }else{
+      rc = SQLITE_OK;
+      id->pLock->cnt = 0;
+    }
   }
+  sqliteOsLeaveMutex();
+  id->locked = 0;
   return rc;
 #endif
 #if OS_WIN
-  return UnlockFile(id, 0, 0, 1024, 0) ? SQLITE_OK : SQLITE_IOERR;
+  int rc;
+  if( id->locked==0 ){
+    rc = SQLITE_OK;
+  }else if( id->locked<0 ){
+    UnlockFile(id->h, 1, 0, MX_LOCKBYTE, 0);
+    rc = SQLITE_OK;
+    id->locked = 0;
+  }else{
+    UnlockFile(id->h, id->locked, 0, 1, 0);
+    rc = SQLITE_OK;
+    id->locked = 0;
+  }
+  return rc;
 #endif
 }
 
@@ -642,13 +766,19 @@ int sqliteOsUnlock(OsFile id){
 */
 int sqliteOsRandomSeed(char *zBuf){
   static int once = 1;
-#if OS_UNIX
+#ifdef SQLITE_TEST
+  /* When testing, always use the same random number sequence.
+  ** This makes the tests repeatable.
+  */
+  memset(zBuf, 0, 256);
+#endif
+#if OS_UNIX && !defined(SQLITE_TEST)
   int pid;
   time((time_t*)zBuf);
   pid = getpid();
   memcpy(&zBuf[sizeof(time_t)], &pid, sizeof(pid));
 #endif
-#if OS_WIN
+#if OS_WIN && !defined(SQLITE_TEST)
   GetSystemTime((LPSYSTEMTIME)zBuf);
 #endif
   if( once ){
@@ -679,6 +809,31 @@ int sqliteOsSleep(int ms){
 #endif
 }
 
+/*
+** Macros used to determine whether or not to use threads.  The
+** SQLITE_UNIX_THREADS macro is defined if we are synchronizing for
+** Posix threads and SQLITE_W32_THREADS is defined if we are
+** synchronizing using Win32 threads.
+*/
+#if OS_UNIX && defined(THREADSAFE) && THREADSAFE
+# include <pthread.h>
+# define SQLITE_UNIX_THREADS 1
+#endif
+#if OS_WIN && defined(THREADSAFE) && THREADSAFE
+# define SQLITE_W32_THREADS 1
+#endif
+
+/*
+** Static variables used for thread synchronization
+*/
+static int inMutex = 0;
+#ifdef SQLITE_UNIX_THREADS
+  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
+#ifdef SQLITE_W32_THREADS
+  static CRITICAL_SECTION cs;
+#endif
+
 /*
 ** The following pair of routine implement mutual exclusion for
 ** multi-threaded processes.  Only a single thread is allowed to
@@ -686,16 +841,34 @@ int sqliteOsSleep(int ms){
 **
 ** SQLite uses only a single Mutex.  There is not much critical
 ** code and what little there is executes quickly and without blocking.
-**
-****** TBD:  The mutex is currently unimplemented.  Until it is
-****** implemented, SQLite is not threadsafe.
 */
-static int inMutex = 0;
 void sqliteOsEnterMutex(){
+#ifdef SQLITE_UNIX_THREADS
+  pthread_mutex_lock(&mutex);
+#endif
+#ifdef SQLITE_W32_THREADS
+  static int isInit = 0;
+  while( !isInit ){
+    static long lock = 0;
+    if( InterlockedIncrement(&lock)==1 ){
+      InitializeCriticalSection(&cs);
+      isInit = 1;
+    }else{
+      Sleep(1);
+    }
+  }
+  EnterCriticalSection(&cs);
+#endif
   assert( !inMutex );
   inMutex = 1;
 }
 void sqliteOsLeaveMutex(){
   assert( inMutex );
   inMutex = 0;
+#ifdef SQLITE_UNIX_THREADS
+  pthread_mutex_unlock(&mutex);
+#endif
+#ifdef SQLITE_W32_THREADS
+  LeaveCriticalSection(&cs);
+#endif
 }
