@@ -11,14 +11,25 @@
 *************************************************************************
 ** A TCL Interface to SQLite
 **
-** $Id: tclsqlite.c,v 1.24 2001/09/28 17:47:14 drh Exp $
+** $Id: tclsqlite.c,v 1.39 2002/07/15 20:58:48 drh Exp $
 */
 #ifndef NO_TCL     /* Omit this whole file if TCL is unavailable */
 
-#include "sqlite.h"
+#include "sqliteInt.h"
 #include "tcl.h"
 #include <stdlib.h>
 #include <string.h>
+#include <assert.h>
+
+/*
+** If TCL uses UTF-8 and SQLite is configured to use iso8859, then we
+** have to do a translation when going between the two.  Set the 
+** UTF_TRANSLATION_NEEDED macro to indicate that we need to do
+** this translation.  
+*/
+#if defined(TCL_UTF_MAX) && !defined(SQLITE_UTF8)
+# define UTF_TRANSLATION_NEEDED 1
+#endif
 
 /*
 ** There is one instance of this structure for each SQLite database
@@ -40,22 +51,19 @@ struct CallbackData {
   Tcl_Interp *interp;       /* The TCL interpreter */
   char *zArray;             /* The array into which data is written */
   Tcl_Obj *pCode;           /* The code to execute for each row */
-  int once;                 /* Set only for the first invocation of callback */
+  int once;                 /* Set for first callback only */
   int tcl_rc;               /* Return code from TCL script */
+  int nColName;             /* Number of entries in the azColName[] array */
+  char **azColName;         /* Column names translated to UTF-8 */
 };
 
-/*
-** If TCL uses UTF-8 and SQLite is configured to use iso8859, then we
-** have to do a translation when going between the two.  Set the 
-** UTF_TRANSLATION_NEEDED macro to indicate that we need to do
-** this translation.  
-*/
-#if defined(TCL_UTF_MAX) && !defined(SQLITE_UTF8)
-# define UTF_TRANSLATION_NEEDED 1
-#endif
-
+#ifdef UTF_TRANSLATION_NEEDED
 /*
 ** Called for each row of the result.
+**
+** This version is used when TCL expects UTF-8 data but the database
+** uses the ISO8859 format.  A translation must occur from ISO8859 into
+** UTF-8.
 */
 static int DbEvalCallback(
   void *clientData,      /* An instance of CallbackData */
@@ -65,50 +73,130 @@ static int DbEvalCallback(
 ){
   CallbackData *cbData = (CallbackData*)clientData;
   int i, rc;
-#ifdef UTF_TRANSLATION_NEEDED
   Tcl_DString dCol;
-#endif
-  if( cbData->zArray[0] ){
-    if( cbData->once ){
+  Tcl_DStringInit(&dCol);
+  if( cbData->azColName==0 ){
+    assert( cbData->once );
+    cbData->once = 0;
+    if( cbData->zArray[0] ){
       Tcl_SetVar2(cbData->interp, cbData->zArray, "*", "", 0);
-      for(i=0; i<nCol; i++){
-        Tcl_SetVar2(cbData->interp, cbData->zArray, "*", azN[i],
-           TCL_LIST_ELEMENT|TCL_APPEND_VALUE);
+    }
+    cbData->azColName = malloc( nCol*sizeof(char*) );
+    if( cbData->azColName==0 ){ return 1; }
+    cbData->nColName = nCol;
+    for(i=0; i<nCol; i++){
+      Tcl_ExternalToUtfDString(NULL, azN[i], -1, &dCol);
+      cbData->azColName[i] = malloc( Tcl_DStringLength(&dCol) + 1 );
+      if( cbData->azColName[i] ){
+        strcpy(cbData->azColName[i], Tcl_DStringValue(&dCol));
+      }else{
+        return 1;
       }
-    }
-    for(i=0; i<nCol; i++){
-      char *z = azCol[i];
-      if( z==0 ) z = "";
-#ifdef UTF_TRANSLATION_NEEDED
-      Tcl_DStringInit(&dCol);
-      Tcl_ExternalToUtfDString(NULL, z, -1, &dCol);
-      Tcl_SetVar2(cbData->interp, cbData->zArray, azN[i], 
-            Tcl_DStringValue(&dCol), 0);
+      if( cbData->zArray[0] ){
+        Tcl_SetVar2(cbData->interp, cbData->zArray, "*",
+             Tcl_DStringValue(&dCol), TCL_LIST_ELEMENT|TCL_APPEND_VALUE);
+        if( azN[nCol]!=0 ){
+          Tcl_DString dType;
+          Tcl_DStringInit(&dType);
+          Tcl_DStringAppend(&dType, "typeof:", -1);
+          Tcl_DStringAppend(&dType, Tcl_DStringValue(&dCol), -1);
+          Tcl_DStringFree(&dCol);
+          Tcl_ExternalToUtfDString(NULL, azN[i+nCol], -1, &dCol);
+          Tcl_SetVar2(cbData->interp, cbData->zArray, 
+               Tcl_DStringValue(&dType), Tcl_DStringValue(&dCol),
+               TCL_LIST_ELEMENT|TCL_APPEND_VALUE);
+          Tcl_DStringFree(&dType);
+        }
+      }
+      
       Tcl_DStringFree(&dCol);
-#else
-      Tcl_SetVar2(cbData->interp, cbData->zArray, azN[i], z, 0);
-#endif
-    }
-  }else{
-    for(i=0; i<nCol; i++){
-      char *z = azCol[i];
-      if( z==0 ) z = "";
-#ifdef UTF_TRANSLATION_NEEDED
-      Tcl_DStringInit(&dCol);
-      Tcl_ExternalToUtfDString(NULL, z, -1, &dCol);
-      Tcl_SetVar(cbData->interp, azN[i], Tcl_DStringValue(&dCol), 0);
-      Tcl_DStringFree(&dCol);
-#else
-      Tcl_SetVar(cbData->interp, azN[i], z, 0);
-#endif
     }
   }
-  cbData->once = 0;
+  if( azCol!=0 ){
+    if( cbData->zArray[0] ){
+      for(i=0; i<nCol; i++){
+        char *z = azCol[i];
+        if( z==0 ) z = "";
+        Tcl_DStringInit(&dCol);
+        Tcl_ExternalToUtfDString(NULL, z, -1, &dCol);
+        Tcl_SetVar2(cbData->interp, cbData->zArray, cbData->azColName[i], 
+              Tcl_DStringValue(&dCol), 0);
+        Tcl_DStringFree(&dCol);
+      }
+    }else{
+      for(i=0; i<nCol; i++){
+        char *z = azCol[i];
+        if( z==0 ) z = "";
+        Tcl_DStringInit(&dCol);
+        Tcl_ExternalToUtfDString(NULL, z, -1, &dCol);
+        Tcl_SetVar(cbData->interp, cbData->azColName[i],
+                   Tcl_DStringValue(&dCol), 0);
+        Tcl_DStringFree(&dCol);
+      }
+    }
+  }
   rc = Tcl_EvalObj(cbData->interp, cbData->pCode);
   if( rc==TCL_CONTINUE ) rc = TCL_OK;
   cbData->tcl_rc = rc;
   return rc!=TCL_OK;
 }
+#endif /* UTF_TRANSLATION_NEEDED */
+
+#ifndef UTF_TRANSLATION_NEEDED
+/*
+** Called for each row of the result.
+**
+** This version is used when either of the following is true:
+**
+**    (1) This version of TCL uses UTF-8 and the data in the
+**        SQLite database is already in the UTF-8 format.
+**
+**    (2) This version of TCL uses ISO8859 and the data in the
+**        SQLite database is already in the ISO8859 format.
+*/
+static int DbEvalCallback(
+  void *clientData,      /* An instance of CallbackData */
+  int nCol,              /* Number of columns in the result */
+  char ** azCol,         /* Data for each column */
+  char ** azN            /* Name for each column */
+){
+  CallbackData *cbData = (CallbackData*)clientData;
+  int i, rc;
+  if( azCol==0 || (cbData->once && cbData->zArray[0]) ){
+    Tcl_SetVar2(cbData->interp, cbData->zArray, "*", "", 0);
+    for(i=0; i<nCol; i++){
+      Tcl_SetVar2(cbData->interp, cbData->zArray, "*", azN[i],
+         TCL_LIST_ELEMENT|TCL_APPEND_VALUE);
+      if( azN[nCol] ){
+        char *z = sqlite_mprintf("typeof:%s", azN[i]);
+        Tcl_SetVar2(cbData->interp, cbData->zArray, z, azN[i+nCol],
+           TCL_LIST_ELEMENT|TCL_APPEND_VALUE);
+        sqlite_freemem(z);
+      }
+    }
+    cbData->once = 0;
+  }
+  if( azCol!=0 ){
+    if( cbData->zArray[0] ){
+      for(i=0; i<nCol; i++){
+        char *z = azCol[i];
+        if( z==0 ) z = "";
+        Tcl_SetVar2(cbData->interp, cbData->zArray, azN[i], z, 0);
+      }
+    }else{
+      for(i=0; i<nCol; i++){
+        char *z = azCol[i];
+        if( z==0 ) z = "";
+        Tcl_SetVar(cbData->interp, azN[i], z, 0);
+      }
+    }
+  }
+  rc = Tcl_EvalObj(cbData->interp, cbData->pCode);
+  if( rc==TCL_CONTINUE ) rc = TCL_OK;
+  cbData->tcl_rc = rc;
+  return rc!=TCL_OK;
+}
+#endif
 
 /*
 ** This is an alternative callback for database queries.  Instead
@@ -124,6 +212,7 @@ static int DbEvalCallback2(
 ){
   Tcl_Obj *pList = (Tcl_Obj*)clientData;
   int i;
+  if( azCol==0 ) return 0;
   for(i=0; i<nCol; i++){
     Tcl_Obj *pElem;
     if( azCol[i] && *azCol[i] ){
@@ -197,22 +286,26 @@ static int DbBusyHandler(void *cd, const char *zTable, int nTries){
 static int DbObjCmd(void *cd, Tcl_Interp *interp, int objc,Tcl_Obj *const*objv){
   SqliteDb *pDb = (SqliteDb*)cd;
   int choice;
-  static char *DB_optStrs[] = {
-     "busy",   "close",  "complete",  "eval",  "timeout", 0
+  static const char *DB_strs[] = {
+    "busy",               "changes",           "close",
+    "complete",           "eval",              "last_insert_rowid",
+    "open_aux_file",      "timeout",           0
   };
-  enum DB_opts {
-     DB_BUSY,  DB_CLOSE, DB_COMPLETE, DB_EVAL, DB_TIMEOUT
+  enum DB_enum {
+    DB_BUSY,              DB_CHANGES,          DB_CLOSE,
+    DB_COMPLETE,          DB_EVAL,             DB_LAST_INSERT_ROWID,
+    DB_OPEN_AUX_FILE,     DB_TIMEOUT,          
   };
 
   if( objc<2 ){
     Tcl_WrongNumArgs(interp, 1, objv, "SUBCOMMAND ...");
     return TCL_ERROR;
   }
-  if( Tcl_GetIndexFromObj(interp, objv[1], DB_optStrs, "option", 0, &choice) ){
+  if( Tcl_GetIndexFromObj(interp, objv[1], DB_strs, "option", 0, &choice) ){
     return TCL_ERROR;
   }
 
-  switch( (enum DB_opts)choice ){
+  switch( (enum DB_enum)choice ){
 
   /*    $db busy ?CALLBACK?
   **
@@ -250,6 +343,25 @@ static int DbObjCmd(void *cd, Tcl_Interp *interp, int objc,Tcl_Obj *const*objv){
     break;
   }
 
+  /*
+  **     $db changes
+  **
+  ** Return the number of rows that were modified, inserted, or deleted by
+  ** the most recent "eval".
+  */
+  case DB_CHANGES: {
+    Tcl_Obj *pResult;
+    int nChange;
+    if( objc!=2 ){
+      Tcl_WrongNumArgs(interp, 2, objv, "");
+      return TCL_ERROR;
+    }
+    nChange = sqlite_changes(pDb->db);
+    pResult = Tcl_GetObjResult(interp);
+    Tcl_SetIntObj(pResult, nChange);
+    break;
+  }
+
   /*    $db close
   **
   ** Shutdown the database
@@ -294,6 +406,7 @@ static int DbObjCmd(void *cd, Tcl_Interp *interp, int objc,Tcl_Obj *const*objv){
     int rc;
 #ifdef UTF_TRANSLATION_NEEDED
     Tcl_DString dSql;
+    int i;
 #endif
 
     if( objc!=5 && objc!=3 ){
@@ -314,6 +427,8 @@ static int DbObjCmd(void *cd, Tcl_Interp *interp, int objc,Tcl_Obj *const*objv){
       cbData.zArray = Tcl_GetStringFromObj(objv[3], 0);
       cbData.pCode = objv[4];
       cbData.tcl_rc = TCL_OK;
+      cbData.nColName = 0;
+      cbData.azColName = 0;
       zErrMsg = 0;
       Tcl_IncrRefCount(objv[3]);
       Tcl_IncrRefCount(objv[4]);
@@ -331,16 +446,72 @@ static int DbObjCmd(void *cd, Tcl_Interp *interp, int objc,Tcl_Obj *const*objv){
       Tcl_SetResult(interp, zErrMsg, TCL_VOLATILE);
       free(zErrMsg);
       rc = TCL_ERROR;
+    }else if( rc!=SQLITE_OK && rc!=SQLITE_ABORT ){
+      Tcl_AppendResult(interp, sqlite_error_string(rc), 0);
+      rc = TCL_ERROR;
     }else{
       rc = cbData.tcl_rc;
     }
     Tcl_DecrRefCount(objv[2]);
 #ifdef UTF_TRANSLATION_NEEDED
     Tcl_DStringFree(&dSql);
+    if( objc==5 && cbData.azColName ){
+      for(i=0; i<cbData.nColName; i++){
+        if( cbData.azColName[i] ) free(cbData.azColName[i]);
+      }
+      free(cbData.azColName);
+      cbData.azColName = 0;
+    }
 #endif
     return rc;
   }
 
+  /*
+  **     $db last_insert_rowid 
+  **
+  ** Return an integer which is the ROWID for the most recent insert.
+  */
+  case DB_LAST_INSERT_ROWID: {
+    Tcl_Obj *pResult;
+    int rowid;
+    if( objc!=2 ){
+      Tcl_WrongNumArgs(interp, 2, objv, "");
+      return TCL_ERROR;
+    }
+    rowid = sqlite_last_insert_rowid(pDb->db);
+    pResult = Tcl_GetObjResult(interp);
+    Tcl_SetIntObj(pResult, rowid);
+    break;
+  }
+
+  /*
+  **     $db open_aux_file  FILENAME
+  **
+  ** Begin using FILENAME as the database file used to store temporary
+  ** tables.
+  */
+  case DB_OPEN_AUX_FILE: {
+    const char *zFilename;
+    char *zErrMsg = 0;
+    int rc;
+    if( objc!=3 ){
+      Tcl_WrongNumArgs(interp, 2, objv, "FILENAME");
+      return TCL_ERROR;
+    }
+    zFilename = Tcl_GetStringFromObj(objv[2], 0);
+    rc = sqlite_open_aux_file(pDb->db, zFilename, &zErrMsg);
+    if( rc!=0 ){
+      if( zErrMsg ){
+        Tcl_AppendResult(interp, zErrMsg, 0);
+        free(zErrMsg);
+      }else{
+        Tcl_AppendResult(interp, sqlite_error_string(rc), 0);
+      }
+      return TCL_ERROR;
+    }
+    break;
+  }
+
   /*
   **     $db timeout MILLESECONDS
   **
@@ -391,6 +562,7 @@ static int DbMain(void *cd, Tcl_Interp *interp, int argc, char **argv){
   int mode;
   SqliteDb *p;
   char *zErrMsg;
+  char zBuf[80];
   if( argc==2 ){
     if( strcmp(argv[1],"-encoding")==0 ){
       Tcl_AppendResult(interp,sqlite_encoding,0);
@@ -430,6 +602,24 @@ static int DbMain(void *cd, Tcl_Interp *interp, int argc, char **argv){
     return TCL_ERROR;
   }
   Tcl_CreateObjCommand(interp, argv[1], DbObjCmd, (char*)p, DbDeleteCmd);
+
+  /* The return value is the value of the sqlite* pointer
+  */
+  sprintf(zBuf, "%p", p->db);
+  if( strncmp(zBuf,"0x",2) ){
+    sprintf(zBuf, "0x%p", p->db);
+  }
+  Tcl_AppendResult(interp, zBuf, 0);
+
+  /* If compiled with SQLITE_TEST turned on, then register the "md5sum"
+  ** SQL function.
+  */
+#ifdef SQLITE_TEST
+  {
+    extern void Md5_Register(sqlite*);
+    Md5_Register(p->db);
+   }
+#endif  
   return TCL_OK;
 }
 
@@ -454,13 +644,13 @@ static int DbMain(void *cd, Tcl_Interp *interp, int argc, char **argv){
 int Sqlite_Init(Tcl_Interp *interp){
   Tcl_InitStubs(interp, "8.0", 0);
   Tcl_CreateCommand(interp, "sqlite", DbMain, 0, 0);
-  Tcl_PkgProvide(interp, "sqlite", "1.0");
+  Tcl_PkgProvide(interp, "sqlite", "2.0");
   return TCL_OK;
 }
 int Tclsqlite_Init(Tcl_Interp *interp){
   Tcl_InitStubs(interp, "8.0", 0);
   Tcl_CreateCommand(interp, "sqlite", DbMain, 0, 0);
-  Tcl_PkgProvide(interp, "sqlite", "1.0");
+  Tcl_PkgProvide(interp, "sqlite", "2.0");
   return TCL_OK;
 }
 int Sqlite_SafeInit(Tcl_Interp *interp){
@@ -535,7 +725,7 @@ int TCLSH_MAIN(int argc, char **argv){
           TCL_GLOBAL_ONLY | TCL_LIST_ELEMENT | TCL_APPEND_VALUE);
     }
     if( Tcl_EvalFile(interp, argv[1])!=TCL_OK ){
-      char *zInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
+      const char *zInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
       if( zInfo==0 ) zInfo = interp->result;
       fprintf(stderr,"%s: %s\n", *argv, zInfo);
       return 1;
