@@ -1,7 +1,7 @@
 #
 # Run this Tcl script to generate the sqlite.html file.
 #
-set rcsid {$Id: sqlite.tcl,v 1.15 2001/09/28 23:11:24 drh Exp $}
+set rcsid {$Id: sqlite.tcl,v 1.18 2002/06/27 13:21:02 drh Exp $}
 
 puts {<html>
 <head>
@@ -12,7 +12,7 @@ puts {<html>
 sqlite: A program to administer SQLite databases
 </h1>}
 puts "<p align=center>
-(This page was last modified on [lrange $rcsid 3 4] GMT)
+(This page was last modified on [lrange $rcsid 3 4] UTC)
 </p>"
 
 puts {
@@ -117,6 +117,14 @@ indices from the database.  You can not make manual changes
 to the sqlite_master table.
 </p>
 
+<p>
+The schema for TEMPORARY tables is not stored in the "sqlite_master" table
+since TEMPORARY tables are not visible to applications other than the
+application that created the table.  The schema for TEMPORARY tables
+is stored in another special table named "sqlite_temp_master".  The
+"sqlite_temp_master" table is temporary itself.
+</p>
+
 <h2>Special commands to sqlite</h2>
 
 <p>
@@ -158,7 +166,7 @@ puts {
 
 <p>The sqlite program is able to show the results of a query
 in five different formats: "line", "column", "list", "html", and "insert".
-You can use the ".mode" dot command to switch between these three output
+You can use the ".mode" dot command to switch between these output
 formats.</p>
 
 <p>The default output mode is "list".  In
@@ -335,8 +343,8 @@ puts {
 executing the following query:</p>
 
 <blockquote><pre>
-SELECT name FROM sqlite_master 
-WHERE type='table' 
+SELECT name FROM sqlite_master WHERE type='table' 
+UNION ALL SELECT name FROM sqlite_temp_master WHERE type='table'
 ORDER BY name;
 </pre></blockquote>
 
@@ -376,7 +384,9 @@ puts {
 list mode, then entering the following query:</p>
 
 <blockquote><pre>
-SELECT sql FROM sqlite_master
+SELECT sql FROM 
+   (SELECT * FROM sqlite_master UNION ALL
+    SELECT * FROM sqlite_temp_master)
 WHERE type!='meta'
 ORDER BY tbl_name, type DESC, name
 </pre></blockquote>
@@ -385,7 +395,9 @@ ORDER BY tbl_name, type DESC, name
 want the schema for a single table, the query looks like this:</p>
 
 <blockquote><pre>
-SELECT sql FROM sqlite_master
+SELECT sql FROM
+   (SELECT * FROM sqlite_master UNION ALL
+    SELECT * FROM sqlite_temp_master)
 WHERE tbl_name LIKE '%s' AND type!='meta'
 ORDER BY type DESC, name
 </pre></blockquote>
